@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2022 Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
+** Copyright (C) 2022-2023 Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -16,7 +16,7 @@
 ** Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 **
 */
-
+#include "config.h"
 
 #include "utils/mu-result.hh"
 #include <array>
@@ -29,9 +29,12 @@
 #include <mu-store.hh>
 #include <mu-maildir.hh>
 #include <utils/mu-utils.hh>
+#include <utils/mu-utils-file.hh>
 #include <utils/mu-test-utils.hh>
 #include <message/mu-message.hh>
 
+#include "mu-query-parser.hh"
+
 using namespace Mu;
 
 
@@ -40,19 +43,20 @@ using TestMap = std::unordered_map<std::string, std::string>;
 
 static Store
 make_test_store(const std::string& test_path, const TestMap& test_map,
-		const StringVec &personal_addresses)
+		Option<const Config&> conf={})
 {
-	std::string maildir = test_path + "/Maildir";
+	const auto maildir{join_paths(test_path, "/Maildir/")};
+	// note the trailing '/'
+	g_test_bug("2513");
 
 	/* write messages to disk */
 	for (auto&& item: test_map) {
 
-		const auto msgpath = maildir + "/" + item.first;
-
 		/* create the directory for the message */
+		const auto msgpath{join_paths(maildir, item.first)};
 		auto dir = to_string_gchar(g_path_get_dirname(msgpath.c_str()));
 		if (g_test_verbose())
-			g_message("create message dir %s", dir.c_str());
+			mu_message("create maildir {}", dir.c_str());
 
 		g_assert_cmpuint(g_mkdir_with_parents(dir.c_str(), 0700), ==, 0);
 
@@ -63,18 +67,11 @@ make_test_store(const std::string& test_path, const TestMap& test_map,
 		stream.close();
 	}
 
-	/* make the store */
-	auto store = Store::make_new(test_path, maildir, personal_addresses, {});
+	auto store = Store::make_new(test_path, maildir, conf);
 	assert_valid_result(store);
 
 	/* index the messages */
-	auto res = store->indexer().start({});
-	g_assert_true(res);
-	while(store->indexer().is_running()) {
-		using namespace std::chrono_literals;
-		std::this_thread::sleep_for(100ms);
-	}
-
+	g_assert_true(store->indexer().start({},true/*block*/));
 	if (test_map.size() > 0)
 		g_assert_false(store->empty());
 
@@ -84,7 +81,6 @@ make_test_store(const std::string& test_path, const TestMap& test_map,
 	return std::move(store.value());
 }
 
-
 static void
 test_simple()
 {
@@ -152,10 +148,15 @@ I said: "Aujourd'hui!"
 			"to:sqlite-dev@sqlite.org",
 			"list:sqlite-dev.sqlite.org",
 			"aujourd'hui",
+#ifdef HAVE_CLD2
+			"lang:en",
+#endif /*HAVE_CLD2*/
 		}) {
 
 		if (g_test_verbose())
-			g_message("query: '%s'", expr);
+			mu_message("query: '{}'\n", expr,
+				make_xapian_query(store, expr)->get_description());
+
 		auto qr = store.run_query(expr);
 		assert_valid_result(qr);
 		g_assert_false(qr->empty());
@@ -544,9 +545,9 @@ Boo!
 	/* create maildir with message */
 	TempDir tdir;
 	auto store{make_test_store(tdir.path(), test_msgs, {})};
-	g_debug("%s", store.properties().root_maildir.c_str());
+	g_debug("%s", store.root_maildir().c_str());
 	/* ensure we have a proper maildir, with new/, cur/ */
-	auto mres = maildir_mkdir(store.properties().root_maildir + "/inbox");
+	auto mres = maildir_mkdir(store.root_maildir() + "/inbox");
 	assert_valid_result(mres);
 	g_assert_cmpuint(store.size(), ==, 1U);
 
@@ -568,11 +569,14 @@ Boo!
 	auto move_opts{rename ? Store::MoveOptions::ChangeName : Store::MoveOptions::None};
 	auto moved_msgs = store.move_message(old_docid, Nothing, Flags::Seen, move_opts);
 	assert_valid_result(moved_msgs);
+
 	g_assert_true(moved_msgs->size() == 1);
-	const auto& moved_msg{moved_msgs->at(0).second};
+	auto&& moved_msg_opt = store.find_message(moved_msgs->at(0).first);
+	g_assert_true(!!moved_msg_opt);
+	const auto&moved_msg = std::move(*moved_msg_opt);
 	const auto new_path = moved_msg.path();
 	if (!rename)
-		assert_equal(new_path, store.properties().root_maildir + "/inbox/cur/msg:2,S");
+		assert_equal(new_path, store.root_maildir() + "/inbox/cur/msg:2,S");
 	g_assert_cmpuint(store.size(), ==, 1);
 	g_assert_false(::access(old_path.c_str(), F_OK) == 0);
 	g_assert_true(::access(new_path.c_str(), F_OK) == 0);
@@ -580,10 +584,9 @@ Boo!
 	/* also ensure that the cached sexp for the message has been updated;
 	 * that's what mu4e uses */
 	const auto moved_sexp{moved_msg.sexp()};
-	//std::cerr << "@@ " << *moved_msg << '\n';
 	g_assert_true(moved_sexp.plistp());
-	g_assert_true(moved_sexp.has_prop(":path"));
-	assert_equal(moved_sexp.get_prop(":path").string(), new_path);
+	g_assert_true(!!moved_sexp.get_prop(":path"));
+	assert_equal(moved_sexp.get_prop(":path").value().string(), new_path);
 
 	/*
 	 * find new message with query, ensure it's really that new one.
@@ -636,7 +639,8 @@ test_term_split()
 	// Note the fancy quote in "foo’s bar"
 	const TestMap test_msgs = {{
 			"inbox/new/msg",
-			{ R"(Message-Id: <abcde@foo.bar>
+			{
+R"(Message-Id: <abcde@foo.bar>
 From: "Foo Example" <bar@example.com>
 Date: Wed, 26 Oct 2022 11:01:54 -0700
 To: example@example.com
@@ -649,17 +653,63 @@ Boo!
 	TempDir tdir;
 	auto store{make_test_store(tdir.path(), test_msgs, {})};
 	/* true: match; false: no match */
-	const auto cases = std::array<std::pair<const char*, bool>, 6>{{
+	const auto cases = std::array<std::pair<const char*, bool>, 8>{{
 		{"subject:foo's", true},
 		{"subject:foo*", true},
 		{"subject:/foo/", true},
 		{"subject:/foo’s/", true}, /* <-- breaks before PR #2365 */
 		{"subject:/foo.*bar/", true},  /* <-- breaks before PR #2365 */
-		{"subject:/foo’s bar/", false}, /* <-- no matching yet */
+		{"subject:/foo’s bar/", false}, /* <-- no matching, needs quoting */
+		{"subject:\"/foo’s bar/\"", true}, /* <-- this works, quote the regex */
+		{R"(subject:"/foo’s bar/")", true}, /* <-- this works, quote the regex */
 	}};
 
 	for (auto&& test: cases) {
-		g_debug("query: %s", test.first);
+		mu_debug("query: '{}'", test.first);
+		auto qr = store.run_query(test.first);
+		assert_valid_result(qr);
+		if (test.second)
+			g_assert_cmpuint(qr->size(), ==, 1);
+		else
+			g_assert_true(qr->empty());
+	}
+}
+
+static void
+test_subject_kata_containers()
+{
+	g_test_bug("2167");
+
+	// Note the fancy quote in "foo’s bar"
+	const TestMap test_msgs = {{
+			"inbox/new/msg",
+			{
+R"(Message-Id: <abcde@foo.bar>
+From: "Foo Example" <bar@example.com>
+Date: Wed, 26 Oct 2022 11:01:54 -0700
+To: example@example.com
+Subject: kata-containers
+
+voodoo-containers
+
+Boo!
+)"},
+		}};
+
+	TempDir tdir;
+	auto store{make_test_store(tdir.path(), test_msgs, {})};
+	/* true: match; false: no match */
+	const auto cases = std::vector<std::pair<const char*, bool>>{{
+			{"subject:kata", true},
+			{"subject:containers", true},
+			{"subject:kata-containers", true},
+			{"subject:\"kata containers\"", true},
+			{"voodoo-containers", true},
+			{"voodoo containers", true}
+		}};
+
+	for (auto&& test: cases) {
+		mu_debug("query: '{}'", test.first);
 		auto qr = store.run_query(test.first);
 		assert_valid_result(qr);
 		if (test.second)
@@ -768,18 +818,74 @@ html
 		assert_valid_result(qr);
 		g_assert_cmpuint(qr->size(), ==, 1);
 	}
-
 }
 
 
+static void
+test_ngrams()
+{
+	g_test_bug("2167");
+
+	// Note the fancy quote in "foo’s bar"
+	const TestMap test_msgs = {{
+			"inbox/new/msg",
+			{
+R"(From: "Bob" <bob@builder.com>
+Subject: スポンサーシップ募集
+To: "Chase" <chase@ppatrol.org>
+Message-Id: 112342343e9dfo.fsf@builder.com
+
+    中文
+
+https://trac.xapian.org/ticket/719
+
+    サーバがダウンしました
+)"}}};
+
+	MemDb mdb;
+	Config conf{mdb};
+	conf.set<Config::Id::SupportNgrams>(true);
+
+	TempDir tdir;
+	auto store{make_test_store(tdir.path(), test_msgs, conf)};
+
+	/* true: match; false: no match */
+	const auto cases = std::vector<std::pair<std::string_view, bool>>{{
+			{"body:中文", true},
+			{"body:中", true},
+			{"body:文", true},
+			{"body:し", true},
+			{"body:サー", true},
+			{"body:サーバがダウンしました", true}, // fail
+			{"中文", true},
+			{"中", true},
+			{"文", true},
+			{"subject:スポン", true },
+			{"subject:スポンサーシップ募集", true },
+			{"subject:シップ", true }, // XXX should match
+			{"サーバがダウンしました", true}, // okay
+			{"body:サーバがダウンしました", true}, //  okay
+			{"subject:スポンサーシップ募集", true}, // okay
+			{"subject:シップx", true }, // XXX should match
+		}};
+
+	for (auto&& test: cases) {
+		auto qr = store.run_query(std::string{test.first});
+		assert_valid_result(qr);
+		if (test.second)
+			g_assert_cmpuint(qr->size(), ==, 1);
+		else
+			g_assert_true(qr->empty());
+	}
+}
+
 int
 main(int argc, char* argv[])
 {
 	mu_test_init(&argc, &argv);
 
-	g_test_bug_base("https://github.com/djcb/mu/issues/");
-
-	g_test_add_func("/store/query/simple", test_simple);
+	g_test_add_func("/store/query/simple",
+			test_simple);
 	g_test_add_func("/store/query/spam-address-components",
 			test_spam_address_components);
 	g_test_add_func("/store/query/dups-related",
@@ -794,9 +900,14 @@ main(int argc, char* argv[])
 			test_duplicate_refresh_rename);
 	g_test_add_func("/store/query/term-split",
 			test_term_split);
+	g_test_add_func("/store/query/kata_containers",
+			test_subject_kata_containers);
 	g_test_add_func("/store/query/related-dup-threaded",
 			test_related_dup_threaded);
-	g_test_add_func("/store/query/html", test_html);
+	g_test_add_func("/store/query/html",
+			test_html);
+	g_test_add_func("/store/query/ngrams",
+			test_ngrams);
 
 	return g_test_run();
 }
